# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/4/13 0:11
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function, division
import io
import ast
import contextlib
from pyeal.exc import *


def _repr(o):
    s = repr(o)
    if isinstance(s, bytes):
        return s.decode('utf-8')
    return s


class Unparser(object):
    def __init__(self, tree, f=None):
        if f is None:
            f = io.StringIO()
        self.f = f
        self.col_off = 0
        # 记录最后一个写入的字符串， 配合_write_code， _write_line方法可以避免不必要的空行
        self.end_char = ''

        self._dispatch(tree)

    def _write_code(self, *s):
        if len(s) > 0:
            for i in s:
                if isinstance(i, bytes):
                    self.f.write(i.decode('utf-8'))
                else:
                    self.f.write(i)
            else:
                if len(i) > 0:
                    self.end_char = i[-1]

    @contextlib.contextmanager
    def _write_block(self):
        self.col_off += 1
        yield
        self.col_off -= 1

    @contextlib.contextmanager
    def _write_line(self):
        if self.end_char == '\n':
            self._write_code('    ' * self.col_off)
        yield
        if self.end_char != '\n':
            self._write_code('\n')

    def _dispatch(self, node):
        if isinstance(node, ast.slice):
            self._dispatch_slice(node)
        elif isinstance(node, ast.stmt):
            self._dispatch_stmt(node)
        elif isinstance(node, ast.expr):
            self._dispatch_expr(node)
        elif isinstance(node, ast.unaryop):
            self._dispatch_unaryop(node)
        elif isinstance(node, ast.boolop):
            self._dispatch_boolop(node)
        elif isinstance(node, ast.operator):
            self._dispatch_operator(node)
        elif isinstance(node, ast.cmpop):
            self._dispatch_cmpop(node)

        elif isinstance(node, ast.mod):
            self._dispatch_mod(node)

        elif isinstance(node, ast.ExceptHandler):
            self._except_handler(node)
        elif isinstance(node, ast.comprehension):
            self._comprehension(node)

        else:
            raise SealException("无法识别的节点: ", node)

    def _dispatch_mod(self, node):
        """
        :type node: ast.mod
        """
        if False:
            pass
        elif isinstance(node, ast.Module):
            self._module(node)
        else:
            raise SealException("无法识别的节点: ", node)

    def _dispatch_slice(self, node):
        """
        :type node: ast.slice
        """
        if isinstance(node, ast.Index):
            self._dispatch(node.value)
        elif isinstance(node, ast.Slice):
            if node.lower is not None:
                self._dispatch(node.lower)
            self._write_code(':')
            if node.upper is not None:
                self._dispatch(node.upper)
            self._write_code(':')
            if node.step is not None:
                self._dispatch(node.step)
        elif isinstance(node, ast.ExtSlice):
            end_id = len(node.dims) - 1
            for id_, i in enumerate(node.dims):
                self._dispatch(i)
                if id_ != end_id:
                    self._write_code(', ')
        else:
            raise SealException("无法识别的节点: ", node)

    def _dispatch_expr(self, node):
        """
        :type node: ast.expr
        """
        if False:
            pass
        elif isinstance(node, ast.Yield):
            if node.value is None:
                self._write_code('yield')
            else:
                self._write_code('yield ')
                self._dispatch(node.value)

        elif isinstance(node, ast.Attribute):
            self._attribute(node)
        elif isinstance(node, ast.Call):
            self._call(node)
        elif isinstance(node, ast.Compare):
            self._compare(node)
        elif isinstance(node, ast.Subscript):
            self._subscript(node)
        elif isinstance(node, ast.IfExp):
            self._if_exp(node)

        elif isinstance(node, ast.UnaryOp):
            self._unary_op(node)
        elif isinstance(node, ast.BoolOp):
            self._bool_op(node)
        elif isinstance(node, ast.BinOp):
            self._bin_op(node)

        elif isinstance(node, ast.Set):
            self._set(node)
        elif isinstance(node, ast.Dict):
            self._dict(node)
        elif isinstance(node, ast.List):
            self._list(node)
        elif isinstance(node, ast.Tuple):
            self._tuple(node)

        elif isinstance(node, ast.DictComp):
            self._dict_comp(node)
        elif isinstance(node, ast.SetComp):
            self._set_comp(node)
        elif isinstance(node, ast.ListComp):
            self._list_comp(node)
        elif isinstance(node, ast.GeneratorExp):
            self._generator_exp(node)

        elif isinstance(node, ast.Str):
            self._str(node)
        elif isinstance(node, ast.Num):
            self._num(node)
        elif isinstance(node, ast.Name):
            self._name(node)
        elif isinstance(node, ast.Lambda):
            self._lambda(node)

        else:
            raise SealException("无法识别的节点: ", node)

    def _dispatch_stmt(self, node):
        """
        :type node: ast.stmt
        """
        if False:
            pass
        elif isinstance(node, ast.ClassDef):
            self._class_def(node)
        elif isinstance(node, ast.FunctionDef):
            self._function_def(node)
        elif isinstance(node, ast.For):
            self._for(node)
        elif isinstance(node, ast.If):
            self._if(node)
        elif isinstance(node, ast.Return):
            self._return(node)
        elif isinstance(node, ast.Pass):
            self._write_code('pass')
        elif isinstance(node, ast.Raise):
            if node.type is None:
                self._write_code('raise')
            else:
                self._write_code('raise ')
                self._dispatch(node.type)
        elif isinstance(node, ast.Assert):
            self._write_code('assert ')
            self._dispatch(node.test)
            if node.msg is not None:
                self._write_code(', ')
                self._dispatch(node.msg)
        elif isinstance(node, ast.Print):
            self._write_code('print ')
            for id_, i in enumerate(node.values):
                if id_ != 0:
                    self._write_code(', ')
                self._dispatch(i)
        elif isinstance(node, ast.Delete):
            self._write_code('del ')
            for id_, i in enumerate(node.targets):
                if id_ != 0:
                    self._write_code(', ')
                self._dispatch(i)

        elif isinstance(node, ast.Global):
            self._write_code('global ', ', '.join(node.names))
        elif isinstance(node, ast.Continue):
            self._write_code('continue')
        elif isinstance(node, ast.Break):
            self._write_code('break')

        elif isinstance(node, ast.AugAssign):
            self._dispatch(node.target)
            self._write_code(' ')
            self._dispatch(node.op)
            self._write_code('= ')
            self._dispatch(node.value)

        elif isinstance(node, ast.Expr):
            self._expr(node)
        elif isinstance(node, ast.ImportFrom):
            self._import_from(node)
        elif isinstance(node, ast.Import):
            self._import(node)
        elif isinstance(node, ast.Assign):
            self._assign(node)

        elif isinstance(node, ast.With):
            self._with(node)
        elif isinstance(node, ast.While):
            self._while(node)
        elif isinstance(node, ast.TryExcept):
            self._try_except(node)
        elif isinstance(node, ast.TryFinally):
            self._try_finally(node)


        else:
            raise SealException("无法识别的节点: ", node)

    def _dispatch_unaryop(self, node):
        typemap = {
            ast.Invert: '~',
            ast.Not: 'not',
            ast.UAdd: '+',
            ast.USub: '-',
        }
        v = typemap.get(type(node), None)
        if v is None:
            raise SealException("无法识别的节点: ", node)
        self._write_code(v)

    def _dispatch_boolop(self, node):
        typemap = {
            ast.And: 'and',
            ast.Or: 'or',
        }
        v = typemap.get(type(node), None)
        if v is None:
            raise SealException("无法识别的节点: ", node)
        self._write_code(v)

    def _dispatch_operator(self, node):
        typemap = {
            ast.Add: '+',
            ast.Sub: '-',
            ast.Mult: '*',
            ast.Div: '/',
            ast.FloorDiv: '//',
            ast.RShift: '>>',
            ast.LShift: '<<',
            ast.BitAnd: '&',
            ast.BitOr: '|',
            ast.BitXor: '^',
            ast.Mod: '%',
            ast.Pow: '**',
        }
        v = typemap.get(type(node), None)
        if v is None:
            raise SealException("无法识别的节点: ", node)
        self._write_code(v)

    def _dispatch_cmpop(self, node):
        """
        :type node: ast.cmpop
        """
        typemap = {
            ast.Is: 'is',
            ast.IsNot: 'is not',
            ast.In: 'in',
            ast.NotIn: 'not in',
            ast.Eq: '==',
            ast.NotEq: '!=',
            ast.Gt: '>',
            ast.GtE: '>=',
            ast.Lt: '<',
            ast.LtE: '<=',
        }
        v = typemap.get(type(node), None)
        if v is None:
            raise SealException("无法识别的节点: ", node)
        self._write_code(v)

    def _module(self, node):
        """
        :type node: ast.Module
        """
        for i in node.body:
            with self._write_line():
                self._dispatch(i)

    def _class_def(self, node):
        """
        :type node: ast.ClassDef
        """
        for i in node.decorator_list:
            with self._write_line():
                self._write_code('@')
                self._dispatch(i)
        # 类头部输出
        with self._write_line():
            self._write_code('class ', node.name, '(')
            end_id = len(node.bases) - 1
            for id_, i in enumerate(node.bases):
                self._dispatch(i)
                if id_ != end_id:
                    self._write_code(', ')
            self._write_code('):')

        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)

    def _function_def(self, node):
        """
        :type node: ast.FunctionDef
        """
        for i in node.decorator_list:
            with self._write_line():
                self._write_code('@')
                self._dispatch(i)
        # 函数头部输出
        with self._write_line():
            self._write_code('def ', node.name, '(')
            end_id = len(node.args.args) - 1
            id_ = 0
            if node.args.vararg is not None:
                end_id += 1
            if node.args.kwarg is not None:
                end_id += 1

            args = node.args.args[:len(node.args.args) - len(node.args.defaults)]
            def_args = node.args.args[len(node.args.args) - len(node.args.defaults):]
            for i in args:
                self._dispatch(i)
                if id_ != end_id:
                    self._write_code(', ')
                id_ += 1
            for i, v in zip(def_args, node.args.defaults):
                self._dispatch(i)
                self._write_code('=')
                self._dispatch(v)
                if id_ != end_id:
                    self._write_code(', ')
                id_ += 1
            if node.args.vararg is not None:
                self._write_code('*', node.args.vararg)
                if id_ != end_id:
                    self._write_code(', ')
                id_ += 1
            if node.args.kwarg is not None:
                self._write_code('**', node.args.kwarg)
                if id_ != end_id:
                    self._write_code(', ')
                id_ += 1
            self._write_code('):')

        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)

    def _for(self, node):
        """
        :type node: ast.For
        """

        # 函数头部输出
        self._write_code('for ')
        self._dispatch(node.target)
        self._write_code(' in ')
        self._dispatch(node.iter)
        self._write_code(':\n')

        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)
        if len(node.orelse) > 0:
            with self._write_line():
                self._write_code('else:')
            with self._write_block():
                for i in node.orelse:
                    with self._write_line():
                        self._dispatch(i)

    def _if(self, node, is_elif=False):
        """
        :type node: ast.If
        :type is_elif: bool
        """

        # 函数头部输出
        if is_elif:
            self._write_code('elif ')
        else:
            self._write_code('if ')
        self._dispatch(node.test)
        self._write_code(':\n')

        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)
        if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If):
            node = node.orelse[0]
            with self._write_line():
                self._if(node, True)
        else:
            if len(node.orelse) > 0:
                with self._write_line():
                    self._write_code('else:\n')
                with self._write_block():
                    for i in node.orelse:
                        with self._write_line():
                            self._dispatch(i)

    def _expr(self, node):
        """
        :type node: ast.Expr
        """
        self._dispatch(node.value)

    def _import_from(self, node):
        """
        :type node: ast.ImportFrom
        """

        co = 'from {}{} import '.format(''.join(('.' for i in range(node.level))),
                                        '' if node.module is None else node.module)
        imps = (i.name if i.asname is None else '{} as {}'.format(i.name, i.asname) for i in node.names)
        imps = ', '.join(imps)
        self._write_code(co, imps)

    def _import(self, node):
        """
        :type node: ast.Import
        """
        imps = (i.name if i.asname is None else '{} as {}'.format(i.name, i.asname) for i in node.names)
        imps = ', '.join(imps)
        self._write_code('import ', imps)

    def _compare(self, node):
        """
        :type node: ast.Compare
        """
        self._dispatch(node.left)
        for o, c in zip(node.ops, node.comparators):
            self._write_code(' ')
            self._dispatch(o)
            self._write_code(' ')
            self._dispatch(c)

    def _assign(self, node):
        """
        :type node: ast.Assign
        """
        self._dispatch(node.targets[0])
        self._write_code(' = ')
        self._dispatch(node.value)

    def _with(self, node):
        """
        :type node: ast.With
        """
        self._write_code('with ')
        self._dispatch(node.context_expr)
        if node.optional_vars is not None:
            self._write_code(' as ')
            self._dispatch(node.optional_vars)
        self._write_code(':\n')
        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)

    def _while(self, node):
        """
        :type node: ast.While
        """
        self._write_code('while ')
        self._dispatch(node.test)
        self._write_code(':\n')
        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)
        if len(node.orelse) > 0:
            with self._write_line():
                self._write_code('else:')
            with self._write_block():
                for i in node.orelse:
                    with self._write_line():
                        self._dispatch(i)

    def _try_except(self, node):
        """
        :type node: ast.TryExcept
        """
        self._write_code('try:\n')
        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)
        for i in node.handlers:
            with self._write_line():
                self._dispatch(i)
        if len(node.orelse) > 0:
            with self._write_line():
                self._write_code('else:')
            with self._write_block():
                for i in node.orelse:
                    with self._write_line():
                        self._dispatch(i)

    def _try_finally(self, node):
        """
        :type node: ast.TryFinally
        """
        if len(node.body) == 1 and isinstance(node.body[0], ast.TryExcept):
            with self._write_line():
                self._dispatch(node.body[0])
        else:
            with self._write_line():
                self._write_code('try:')
            with self._write_block():
                for i in node.body:
                    with self._write_line():
                        self._dispatch(i)
        with self._write_line():
            self._write_code('finally:')
        with self._write_block():
            for i in node.finalbody:
                with self._write_line():
                    self._dispatch(i)

    def _except_handler(self, node):
        """
        :type node: ast.ExceptHandler
        """
        if node.name is None:
            self._write_code('except:\n')
        elif node.type is None:
            self._write_code('except ')
            self._dispatch(node.name)
            self._write_code(':\n')
        else:
            self._write_code('except ')
            self._dispatch(node.name)
            self._write_code(' as ')
            self._dispatch(node.type)
            self._write_code(':\n')
        with self._write_block():
            for i in node.body:
                with self._write_line():
                    self._dispatch(i)

    def _return(self, node):
        """
        :type node: ast.Return
        """
        if node.value is None:
            self._write_code('return')
        else:
            self._write_code('return ')
            self._dispatch(node.value)

    def _call(self, node):
        """
        :type node: ast.Call
        """
        self._dispatch(node.func)
        self._write_code('(')
        id_ = 0
        end_id = len(node.args) + len(node.keywords) - 1
        if node.starargs is not None:
            end_id += 1
        if node.kwargs is not None:
            end_id += 1

        for i in node.args:
            self._dispatch(i)
            if id_ != end_id:
                self._write_code(', ')
                id_ += 1
        for i in node.keywords:
            self._write_code(i.arg, '=')
            self._dispatch(i.value)
            if id_ != end_id:
                self._write_code(', ')
                id_ += 1
        if node.starargs is not None:
            self._write_code('*')
            self._dispatch(node.starargs)
            if id_ != end_id:
                self._write_code(', ')
                id_ += 1
        if node.kwargs is not None:
            self._write_code('**')
            self._dispatch(node.kwargs)
            if id_ != end_id:
                self._write_code(', ')
                id_ += 1
        self._write_code(')')

    def _attribute(self, node):
        """
        :type node: ast.Attribute
        """
        self._dispatch(node.value)
        self._write_code('.', node.attr)

    def _subscript(self, node):
        """
        :type node: ast.Subscript
        """

        self._dispatch(node.value)
        self._write_code('[')
        self._dispatch(node.slice)
        self._write_code(']')

    def _if_exp(self, node):
        """
        :type node: ast.IfExp
        """
        self._dispatch(node.body)
        self._write_code(' if ')
        self._dispatch(node.test)
        self._write_code(' else ')
        self._dispatch(node.orelse)

    def _unary_op(self, node):
        """
        :type node: ast.UnaryOp
        """
        self._dispatch(node.op)
        self._write_code(' ')
        self._dispatch(node.operand)

    def _bool_op(self, node):
        """
        :type node: ast.BoolOp
        """
        for id_, i in enumerate(node.values):
            if id_ != 0:
                self._write_code(' ')
                self._dispatch(node.op)
                self._write_code(' ')
            self._dispatch(i)

    def _bin_op(self, node):
        """
        :type node: ast.BinOp
        """
        self._dispatch(node.left)
        self._write_code(' ')
        self._dispatch(node.op)
        self._write_code(' ')
        self._dispatch(node.right)

    def _set(self, node):
        """
        :type node: ast.Set
        """
        self._write_code('{')
        if len(node.elts) > 0:
            for i in node.elts:
                self._dispatch(i)
                self._write_code(', ')
        self._write_code('}')

    def _dict(self, node):
        """
        :type node: ast.Dict
        """
        self._write_code('{')
        if len(node.keys) > 0:
            for k, v in zip(node.keys, node.values):
                self._dispatch(k)
                self._write_code(': ')
                self._dispatch(v)
                self._write_code(', ')
        self._write_code('}')

    def _list(self, node):
        """
        :type node: ast.List
        """
        self._write_code('[')
        if len(node.elts) > 0:
            for i in node.elts:
                self._dispatch(i)
                self._write_code(', ')
        self._write_code(']')

    def _tuple(self, node):
        """
        :type node: ast.Tuple
        """
        self._write_code('(')
        if len(node.elts) > 0:
            for i in node.elts:
                self._dispatch(i)
                self._write_code(', ')
        self._write_code(')')

    def _dict_comp(self, node):
        """
        :type node: ast.DictComp
        """
        self._write_code('{')
        self._dispatch(node.key)
        self._write_code(': ')
        self._dispatch(node.value)
        for i in node.generators:
            self._write_code(' ')
            self._dispatch(i)
        self._write_code('}')

    def _set_comp(self, node):
        """
        :type node: ast.SetComp
        """
        self._write_code('{')
        self._dispatch(node.elt)
        for i in node.generators:
            self._write_code(' ')
            self._dispatch(i)
        self._write_code('}')

    def _list_comp(self, node):
        """
        :type node: ast.ListComp
        """
        self._write_code('[')
        self._dispatch(node.elt)
        for i in node.generators:
            self._write_code(' ')
            self._dispatch(i)
        self._write_code(']')

    def _generator_exp(self, node):
        """
        :type node: ast.GeneratorExp
        """
        self._write_code('(')
        self._dispatch(node.elt)
        for i in node.generators:
            self._write_code(' ')
            self._dispatch(i)
        self._write_code(')')

    def _comprehension(self, node):
        """
        :type node: ast.comprehension
        """
        self._write_code('for ')
        self._dispatch(node.target)
        self._write_code(' in ')
        self._dispatch(node.iter)
        for i in node.ifs:
            self._write_code(' if ')
            self._dispatch(i)

    def _str(self, node):
        """
        :type node: ast.Str
        """
        self._write_code(repr(node.s))
        # co = _repr(node.s)
        # if co[0] == 'b' or co[0] == 'u':
        #     co = co[1:]
        # if isinstance(co, bytes):
        #     self._write_code('b', co)
        # else:
        #     self._write_code('u', co)

    def _num(self, node):
        """
        :type node: ast.Num
        """
        self._write_code(repr(node.n))

    def _name(self, node):
        """
        :type node: ast.Name
        """
        self._write_code(node.id)

    def _lambda(self, node):
        """
        :type node: ast.Lambda
        """
        self._write_code('lambda ')
        end_id = len(node.args.args) - 1
        id_ = 0
        if node.args.vararg is not None:
            end_id += 1
        if node.args.kwarg is not None:
            end_id += 1

        args = node.args.args[:len(node.args.args) - len(node.args.defaults)]
        def_args = node.args.args[len(node.args.args) - len(node.args.defaults):]
        for i in args:
            self._dispatch(i)
            if id_ != end_id:
                self._write_code(', ')
            id_ += 1
        for i, v in zip(def_args, node.args.defaults):
            self._dispatch(i)
            self._write_code('=')
            self._dispatch(v)
            if id_ != end_id:
                self._write_code(', ')
            id_ += 1
        if node.args.vararg is not None:
            self._write_code('*', node.args.vararg)
            if id_ != end_id:
                self._write_code(', ')
            id_ += 1
        if node.args.kwarg is not None:
            self._write_code('**', node.args.kwarg)
            if id_ != end_id:
                self._write_code(', ')
            id_ += 1
        self._write_code(':')

        self._dispatch(node.body)


def parser(code):
    """
    :type code: bytes
    :rtype: str
    """
    return ast.parse(code)


def unparse(tree):
    f = io.StringIO()
    Unparser(tree, f)
    return f.getvalue()



# -*-coding:utf-8 -*-
u"""
:创建时间: 2021/12/19 10:44
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function

import uuid
from abc import abstractmethod

import ast
import os
import locale

from pyeal.code.core import Code

from pyeal.res import BaseRes

from pyeal.module_data import ModuleData


class BuilderBase(object):
    def __init__(self, source, target):
        """
        :type source: BaseRes
        :type target: BaseRes
        """
        self.source = source
        self.target = target

    @abstractmethod
    def build(self):
        pass


class EncapsulationBuilder(BuilderBase):
    """打包编译器"""

    def __init__(self, source, target, name, code, imp_name=None):
        """
        :type source: BaseRes
        :type target: BaseRes
        :type name: unicode
        """
        super(EncapsulationBuilder, self).__init__(source, target)
        self.name = name
        self.code = code
        if imp_name is None:
            self.imp_name = name
        else:
            self.imp_name = imp_name
        self.module_data = ModuleData(self.source)
        self.uid = uuid.uuid4().hex[0:4]

    def seal_name(self):
        return "{}_{}".format(self.name, self.uid)

    def target_name(self, p):
        return "{}_{}".format(self.seal_name(), p)

    def target_path(self, p):
        if len(p.split(self.source.sep())) == 1 and \
                p.split('.')[-1] != 'py':
            return p
        else:
            return "{}_{}".format(self.seal_name(), p)

    def compile_import_node(self, n, m):
        alias_list = []
        for i in n.names:
            if m is None:
                r_find_name = None
            else:
                r_find_name = self.module_data.relative_find_module(i.name, m)
            name = self.module_data.find_module(i.name)
            if r_find_name is None and name is not None:
                if i.asname is None:
                    alias_list.append(ast.alias(asname=i.name.split(".")[0],
                                                name=self.target_name(i.name.split('.')[0])))
                    alias_list.append(ast.alias(asname="_",
                                                name=self.target_name(name)))
                else:
                    target_asname = i.asname
                    alias_list.append(ast.alias(asname="_",
                                                name=self.target_name(i.name.split('.')[0])))
                    alias_list.append(ast.alias(asname=target_asname,
                                                name=self.target_name(name)))

            else:
                alias_list.append(i)
        return [ast.Import(names=[i]) for i in alias_list]

    def compile_import_from_node(self, n, m):
        if n.level == 0:
            nodes = []
            if m is None:
                r_find_name = None
            else:
                r_find_name = self.module_data.relative_find_module(n.module, m)
            name = self.module_data.find_module(n.module)
            if r_find_name is None and name is not None:
                nodes.append(ast.ImportFrom(level=0,
                                            module=self.target_name(name),
                                            names=n.names))
            else:
                nodes.append(n)
            return nodes
        else:
            return [n]

    def delete_future(self, n):
        if isinstance(n, ast.ImportFrom):
            if n.level == 0:
                if n.module == "__future__":
                    return True
        return False

    def compile_module(self, code, m):
        """
        :type code: bytes
        :type m: unicode|None
        :return:
        """
        code_editor = Code(code)

        code_editor.replace_node(
            check_key=lambda n: isinstance(n, ast.Import),
            key=lambda n: self.compile_import_node(n, m)
        ).replace_node(
            check_key=lambda n: isinstance(n, ast.ImportFrom),
            key=lambda n: self.compile_import_from_node(n, m)
        )

        deleted_future = code_editor.delete_node(key=self.delete_future)

        code_editor.insert_to_head(deleted_future)

        code = code_editor.unparse()

        return code.encode("utf-8")

    def build(self):
        for f in self.source.files():
            self.target.write(self.target_path(f), self.source.read(f))
        for m, f in self.module_data.module_name_and_paths():
            code = self.source.read(f)
            code = self.compile_module(code, m)
            self.target.write(self.target_path(f), code)
        code = self.compile_module(self.code, None)
        self.target.write(self.imp_name + ".py", code)


class InstallBuilder(BuilderBase):
    PATH = os.path.dirname(os.path.abspath(__file__))
    with open(os.sep.join((PATH, "mel_template_lib.mel")), "rb") as f:
        mel_template_lib_code = f.read().decode("utf-8")
    mel_template = r'''startInstall(
    "exec(compile(open(plugin_path+" + <<exec_file_name>> + ",'rb').read(), plugin_path+" + <<exec_file_name>> + ", 'exec'), globals(), locals())", 
    <<ann>>, 
    "dist/log.ico", 
    <<plugin_path>>
    );'''

    def __init__(self, source, target, log, ann, name, install_mel_name="install.mel", exec_file_name="exec.py"):
        super(InstallBuilder, self).__init__(source, target)
        self.install_mel_name = install_mel_name
        self.exec_file_name = exec_file_name
        self.log = log
        self.ann = ann
        self.name = name

    def build(self):
        plugin_path = "dist/plugin"
        for f in self.source.files():
            self.target.write(self.target.sep().join((plugin_path, f)), self.source.read(f))
        t = self.mel_template
        t = t.replace("<<exec_file_name>>", '"\'{}.py\'"'.format(self.name))
        t = t.replace("<<ann>>", '"{}"'.format(self.ann))
        t = t.replace("<<plugin_path>>", '"{}/"'.format(plugin_path))
        # 以系统编码写入安装文件
        self.target.write(self.install_mel_name, (self.mel_template_lib_code + t).encode(locale.getpreferredencoding()))
        self.target.write(self.target.sep().join(("dist", "log.ico")), self.log)


if __name__ == "__main__":
    from pyeal.res import LocalRes, DirectoryRes

    root = r"D:\backup_to_cloud\dev\python_for_maya\package\seal\test"
    src = LocalRes(root + r"\src")
    build = LocalRes(root + r"\build")
    mid = DirectoryRes(build, "mid")
    d0 = DirectoryRes(mid, "0")
    dist = DirectoryRes(build, "dist")
    EncapsulationBuilder(src,
                         d0,
                         "test_seal_name",
                         u'''import main
from main import main
main()''').build()
    InstallBuilder(d0, dist, "", "这是一段注释~").build()



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/3/8 12:02
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function, division

import datetime
import json
import os.path
import sys
import uuid
from collections import OrderedDict

from pyeal.res import LocalRes, DirectoryRes, MergeRes
from pyeal.core import EncapsulationBuilder, InstallBuilder
from pyeal.exc import *
from pyeal._command import call_command

if False:
    from typing import List, Tuple, Dict, AnyStr, Any, Callable, Optional

PATH = os.path.dirname(os.path.abspath(__file__))


class Config(object):
    def __init__(self, root):
        root = os.path.abspath(root)
        self.root_path = os.path.abspath(root)
        self.src_path = os.sep.join((self.root_path, 'src'))
        self.lib_path = os.sep.join((self.root_path, 'lib'))
        self.build_path = os.sep.join((self.root_path, 'build'))
        self.middle_path = os.sep.join((root, 'build', 'middle'))
        self.out_path = os.sep.join((root, 'build', 'out'))

        self.root = LocalRes(self.root_path)
        self.src = LocalRes(self.src_path)
        self.lib = LocalRes(self.lib_path)
        self.build = LocalRes(self.build_path)
        self.middle = LocalRes(self.middle_path)
        self.out = LocalRes(self.out_path)

        self.config = json.loads(self.root.read_string("pyeal.json"))
        self.type = self.config["type"]
        self.name = self.config["name"]

    def get_script(self):
        script = self.config.get("exec_script", None)
        if script is None:
            index_module = self.config.get("index_module", "index.py")
            script = self.root.read(index_module)
        return script

    def get_icon(self):
        return self.root.read('icon.ico')

    def get(self, key, default=None):
        return self.config.get(key, default)

    def ass(self, key):
        c = self.get(key, None)
        if c is None:
            raise ConfigException("需要名为<{}>的配置项".format(key))
        return c


def target_is_exec(config):
    EncapsulationBuilder(
        MergeRes(config.src, config.lib),
        config.out,
        config.name,
        config.get_script(),
        config.get('imp_name', config.name),
    ).build()


def target_is_maya_plugin(config):
    m0 = DirectoryRes(config.middle, "m0")

    EncapsulationBuilder(
        MergeRes(config.src, config.lib),
        m0,
        config.name,
        config.get_script(),
        config.get('imp_name', config.name)
    ).build()

    InstallBuilder(
        m0,
        config.out,
        log=config.get_icon(),
        ann=config.get("annotation", "这个插件作者很懒没有写注释哦~"),
        name=config.get('imp_name', config.name),
    ).build()


def target_is_template(config):
    template = DirectoryRes(config.root, 'template')
    template_output = DirectoryRes(config.out, config.ass('template-output'))

    for f in template.files():
        config.out.write(f, template.read(f))

    EncapsulationBuilder(
        MergeRes(config.src, config.lib),
        template_output,
        config.name,
        config.get_script(),
        config.get('imp_name', config.name),
    ).build()


target_types = {
    'package': target_is_exec,
    'maya-plugin': target_is_maya_plugin,
    'template': target_is_template,
}  # type: Dict[AnyStr, Callable[[Config], None]]


def _call_command_list(command_list, build_data):
    for i in command_list:
        if isinstance(i, list):
            call_command(i, build_data)
        elif isinstance(i, dict):
            try:
                is_do_command = eval(i.get('if_expression', 'True'), dict(), build_data)
            except:
                is_do_command = False
            if is_do_command:
                command = i.get('true_command', None)
                if command:
                    call_command(command, build_data)
            else:
                command = i.get('false_command', None)
                if command:
                    call_command(command, build_data)
        else:
            raise SealException("未知的命令格式")


def cmd_build(config, *args):
    """
    :type config: Config
    """
    config.build.clean()
    command_list_at_start = config.get('command_list_at_start', [])
    command_list_at_end = config.get('command_list_at_end', [])
    build_data = {
        'build_uuid': uuid.uuid4().hex,
        'build_time': datetime.datetime.now(),
        'src_path': config.src_path,
        'lib_path': config.lib_path,
        'build_path': config.build_path,
        'middle_path': config.middle_path,
        'out_path': config.out_path,
        'args': args,
        'name': config.name,
    }

    _call_command_list(command_list_at_start, build_data)

    target_type_func = target_types.get(config.type)

    if target_type_func is None:
        raise SealException("未知编译类型")

    target_type_func(config)

    _call_command_list(command_list_at_end, build_data)


config_templates = {
    "maya-plugin": OrderedDict([
        ("type", "maya-plugin"),
        ("source", "src"),
        ("target", "build"),
        ("name", "your-name"),
        ("exec_script", "your-script"),
    ])
}


def cmd_init(config):
    """
    :type config: Config
    """
    config.root.write_string("pyeal.json", json.dumps(config_templates["maya-plugin"], indent=2).encode("utf-8"))
    with open(os.sep.join((PATH, "log.ico")), "rb") as f:
        config.root.write("log.ico", f.read())
    src = os.sep.join((config.root_path, "src"))
    if not os.path.isdir(src):
        os.makedirs(src)
    build = os.sep.join((config.root_path, "build"))
    if not os.path.isdir(build):
        os.makedirs(build)


def cmd_clean(config):
    """
    :type config: Config
    """
    config.build.clean()


commands = {
    "build": cmd_build,
    "init": cmd_init,
    "clean": cmd_clean,
}  # type: Dict[AnyStr, Callable[[Config, ...], None]]


def main():
    root = os.path.abspath(".")
    command = commands.get(sys.argv[1])
    if command is None:
        raise SealException("未知指令")
    command(Config(root), *sys.argv[2:])



# -*-coding:utf-8 -*-
u"""
:创建时间: 2022/4/21 21:55
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function
import ast
import astunparse


# import pyeal.code.unparse as astunparse

def _edit_node_blocks(rn, key):
    if hasattr(rn, "_fields"):
        for i in rn._fields:
            body = getattr(rn, i)
            if isinstance(body, list):
                new_body = key(body)
                for n in new_body:
                    _edit_node_blocks(n, key)
                setattr(rn, i, new_body)


class Code(object):
    def __init__(self, code):
        self.tree = ast.parse(code)

    def insert_to_head(self, *nodes):
        for i in reversed(nodes):
            self.tree.body.insert(0, i)
        return self

    def replace_node(self, check_key, key):
        def fn(body):
            new_body = []
            for n in body:
                if check_key(n):
                    new_body.append(key(n))
                else:
                    new_body.append([n])
            return [ii for i in new_body for ii in i]

        _edit_node_blocks(self.tree, fn)
        return self

    def delete_node(self, key):
        deleted_node = []

        def fn(n):
            deleted_node.append(n)
            return []

        self.replace_node(check_key=key, key=fn)
        return deleted_node

    def unparse(self):
        return astunparse.unparse(self.tree)



# -*-coding:utf-8 -*-
u"""
:创建时间: 2021/12/22 7:38
:作者: 苍之幻灵
:我的主页: https://cpcgskill.com
:QQ: 2921251087
:爱发电: https://afdian.net/@Phantom_of_the_Cang
:aboutcg: https://www.aboutcg.org/teacher/54335
:bilibili: https://space.bilibili.com/351598127

"""
from __future__ import unicode_literals, print_function

import os
from abc import abstractmethod

from pyeal.exc import *


class BaseRes(object):
    @abstractmethod
    def read(self, p):
        """
        :type p: unicode
        :rtype: bytes
        """

    @abstractmethod
    def write(self, p, data):
        """
        :type p: unicode
        :type data: bytes
        :return:
        """

    def read_string(self, p):
        return self.read(p).decode('utf-8')

    def write_string(self, p, s):
        self.write(p, s.encode('utf-8'))

    @abstractmethod
    def clean(self, p=None):
        """
        :rtype:
        """

    @abstractmethod
    def sep(self):
        """
        :rtype: unicode
        """

    @abstractmethod
    def walk(self):
        """

        :rtype: list[tuple[unicode|None, list[unicode]]]
        """

    def files(self):
        for r, fs in self.walk():
            for f in fs:
                if r is not None:
                    f = self.sep().join((r, f))
                yield f


class LocalRes(BaseRes):
    def __init__(self, root):
        """

        :type root: unicode
        """
        self.root = root

    def read(self, p):
        p = self.sep().join((self.root, p))
        with open(p, "rb") as f:
            return f.read()

    def write(self, p, data):
        p = self.sep().join((self.root, p))
        dir_, name = os.path.split(p)
        if not os.path.isdir(dir_):
            os.makedirs(dir_)
        with open(p, "wb") as f:
            return f.write(data)

    def clean(self, p=None):
        file_list = []
        dir_list = []
        root = self.root if p is None else self.sep().join([self.root, p])
        for root, ds, fs in os.walk(root):
            for i in fs:
                file_list.append(self.sep().join([root, i]))
            for i in ds:
                dir_list.append(self.sep().join([root, i]))
        for i in file_list:
            os.remove(i)
        for i in reversed(dir_list):
            os.rmdir(i)

    def sep(self):
        return os.sep

    def walk(self):
        for root, _, files in os.walk(self.root):
            root = self.sep().join(root.split(self.sep())[len(self.root.split(self.sep())):])
            yield None if root == "" else root, files


class DirectoryRes(BaseRes):
    def __init__(self, res, dir_):
        """

        :type res: BaseRes
        :type dir_: unicode
        """
        self.res = res
        self.dir = dir_

    def new_path(self, p):
        return self.sep().join((self.dir, p))

    def read(self, p):
        return self.res.read(self.new_path(p))

    def write(self, p, data):
        return self.res.write(self.new_path(p), data)

    def sep(self):
        return self.res.sep()

    def clean(self, p=None):
        p = self.dir if p is None else self.new_path(p)
        self.res.clean(p)

    def walk(self):
        for r, fs in self.res.walk():
            if r is not None and self.dir in r:
                rs = r.split(self.sep())
                if self.dir == rs[0]:
                    r = self.sep().join(rs[1:])
                    yield None if r == "" else r, fs


class MergeRes(BaseRes):
    def __init__(self, *res):
        """

        :type res: BaseRes
        """
        self.res = res

    def read(self, p):
        for r in self.res:
            try:
                return r.read(p)
            except:
                pass
        raise ResException("找不到对应文件")

    def write(self, p, data):
        return self.res[0].write(p, data)

    def sep(self):
        return self.res[0].sep()

    def clean(self, p=None):
        for r in self.res:
            r.clean(p)

    def walk(self):
        for r in self.res:
            for r, fs in r.walk():
                yield r, fs


if __name__ == "__main__":
    for i in LocalRes(r"D:\backup_to_cloud\dev\python_for_maya\package\seal").walk():
        print(i)
